import math, random

import gym
import numpy as np

import sys
import torch
import torch.nn as nn
import torch.optim as optim
import torch.autograd as autograd 
import torch.nn.functional as F

from IPython.display import clear_output
import matplotlib.pyplot as plt

from collections import deque

import argparse

assert torch.cuda.is_available()
device = torch.device("cuda" if torch.cuda.is_available() else "cpu" )

ENV_ID = "CartPole-v1"
NUM_SEEDS = 10

EPSILON_START = 1.0
EPSILON_FINAL = 0.01
EPSILON_DECAY = 500

NUM_EPISODES = 100
BATCH_SIZE = 64
GAMMA = 0.98

ACTF_FN = 'ReLU'
LR = 0.001
OPTIM = 'Adam'

class ReplayBuffer(object):
    def __init__(self, capacity):
        self.buffer = deque(maxlen=capacity)
    
    def push(self, state, action, reward, next_state, done):
        state      = np.expand_dims(state, 0)
        next_state = np.expand_dims(next_state, 0)
            
        self.buffer.append((state, action, reward, next_state, done))
    
    def sample(self, batch_size):
        state, action, reward, next_state, done = zip(*random.sample(self.buffer, batch_size))
        return np.concatenate(state), action, reward, np.concatenate(next_state), done
    
    def __len__(self):
        return len(self.buffer)


class DQN(nn.Module):
    def __init__(self, num_inputs, num_actions, actf_fn):
        super(DQN, self).__init__()
        if actf_fn == 'ReLU':
            self.actf_fn = nn.ReLU
        elif actf_fn == 'LeakyReLU':
            self.actf_fn = nn.LeakyReLU
        else:
            raise ValueError("Bad argument for actf-fn")
        
        self.layers = nn.Sequential(
            nn.Linear(num_inputs, 128),
            self.actf_fn(),
            nn.Linear(128, 128),
            self.actf_fn(),
            nn.Linear(128, num_actions)
        )
        
    def forward(self, x):
        return self.layers(x)
    
    @torch.no_grad()
    def act(self, state, epsilon):
        if random.random() > epsilon:
            state   = torch.FloatTensor(state).unsqueeze(0).to(device)
            q_value = self.forward(state)
            action  = q_value.max(1)[1].data[0].item()
        else:
            action = random.randrange(env.action_space.n)
        return action

def update_target(current_model, target_model):
    target_model.load_state_dict(current_model.state_dict())

def compute_td_loss(batch_size):
    state, action, reward, next_state, done = replay_buffer.sample(batch_size)

    state      = torch.FloatTensor(np.float32(state)).to(device)
    next_state = torch.FloatTensor(np.float32(next_state)).to(device)
    action     = torch.LongTensor(action).to(device)
    reward     = torch.FloatTensor(reward).to(device)
    done       = torch.BoolTensor(done).to(device)

    q_values      = current_model(state)
    q_value       = q_values.gather(1, action.unsqueeze(1)).squeeze(1) 
    
    with torch.no_grad():
        next_q_values = current_model(next_state)
        next_q_state_values = target_model(next_state) 
        next_q_value = next_q_state_values.gather(1, torch.max(next_q_values, 1)[1].unsqueeze(1)).squeeze(1)
        next_q_value[done] = 0.0
        next_q_value = next_q_value.detach()

    expected_q_value = reward + GAMMA * next_q_value
    
    loss = (q_value - expected_q_value.data).pow(2).mean()
        
    optimizer.zero_grad()
    loss.backward()
    optimizer.step()
    
    return loss

def plot(loss_matrix, reward_matrix, error_bars=False):
    plt.style.use('ggplot')
    loss_matrix = np.array(loss_matrix)
    max_episodes = max([len(rewards) for rewards in reward_matrix])
    #reward_matrix = np.array([rewards[:min_episodes] for rewards in reward_matrix])
    reward_matrix = np.array([(rewards + [rewards[-1]]*(max_episodes - len(rewards))) for rewards in reward_matrix])

    plt.figure(figsize=(20,5))
    plt.subplot(131)
    plt.title('rewards')

    reward_matrix_mean = np.mean(reward_matrix, axis=0)
    reward_matrix_percentiles = np.percentile(reward_matrix, [5, 25, 50, 75, 95], axis=0)

    plt.plot(reward_matrix_mean, linewidth=2, color='green', alpha=0.8)
    if error_bars:
        plt.fill_between([x for x in range(reward_matrix.shape[1])], reward_matrix_percentiles[1, :], reward_matrix_percentiles[3, :], color='#888888', alpha=0.4)
        plt.fill_between([x for x in range(reward_matrix.shape[1])], reward_matrix_percentiles[0, :], reward_matrix_percentiles[4, :], color='#888888', alpha=0.2)
    else:
        for i in range(reward_matrix.shape[0]):
            plt.plot(reward_matrix[i, :], color='gray', linestyle='dashed', alpha=0.5)
    
    plt.subplot(132)
    plt.title('loss')

    loss_matrix_mean = np.mean(loss_matrix, axis=0)
    loss_matrix_percentiles = np.percentile(loss_matrix, [5, 25, 50, 75, 95], axis=0)

    plt.plot(loss_matrix_mean, linewidth=2, alpha=0.8)
    plt.fill_between([x for x in range(loss_matrix.shape[1])], loss_matrix_percentiles[1, :], loss_matrix_percentiles[3, :], color='#888888', alpha=0.4)
    plt.fill_between([x for x in range(loss_matrix.shape[1])], loss_matrix_percentiles[0, :], loss_matrix_percentiles[4, :], color='#888888', alpha=0.2)

    plt.savefig('./imgs/dqn_bp_seeds={}_actf_fn={}_lr={}_optim={}.png'.format(NUM_SEEDS, ACTF_FN, LR, OPTIM))

def plot2(reward_matrix):
    plt.style.use('ggplot')
    max_average_reward = np.max(np.mean(reward_matrix, axis=0))
    plt.figure(figsize=(10, 5))
    plt.title('CartPole-v1 : Vanilla DQN with BP (Max Avg Reward = {:.2f})'.format(max_average_reward))
    reward_matrix_mean = np.mean(reward_matrix, axis=0)
    plt.plot(reward_matrix_mean, linewidth=2, color='blueviolet', alpha=0.8)

    upper_confidence_interval = []
    lower_confidence_interval = []

    for episode in reward_matrix.transpose():
        bootstrapped_means = []
        for B in range(2000):
            bootstrapped_sample = np.random.choice(episode, len(episode))
            bootstrapped_means.append(np.mean(bootstrapped_sample))
        bootstrapped_percentiles = np.percentile(np.array(bootstrapped_means), [2.5, 97.5])
        upper_confidence_interval.append(bootstrapped_percentiles[1])
        lower_confidence_interval.append(bootstrapped_percentiles[0])

    plt.ylim([0, 500])
    plt.xlabel('episodes')
    plt.ylabel('reward')
    plt.fill_between([x for x in range(reward_matrix.shape[1])], upper_confidence_interval, lower_confidence_interval, color='blueviolet', alpha=0.4)
    plt.savefig('./imgs/dqn_bp_rewards_seeds={}_actf_fn={}_lr={}_optim={}.png'.format(NUM_SEEDS, ACTF_FN, LR, OPTIM))
        
def save_data(reward_matrix):
    max_average_reward = np.max(np.mean(reward_matrix, axis=0))

    with open('./data/dqn_bp_seeds={}_actf_fn={}_lr={}_optim={}.txt'.format(NUM_SEEDS, ACTF_FN, LR, OPTIM), 'w') as outfile:
        outfile.write('Max Average Reward : {:.2f}'.format(max_average_reward))
    print('actf_fn : {}, lr : {}, optim : {}'.format(ACTF_FN, LR, OPTIM))
    print('Experiment done : Max Average Reward : {:.2f}'.format(max_average_reward))

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--n-seed", default=NUM_SEEDS)
    parser.add_argument("--actf-fn", default=ACTF_FN)
    parser.add_argument("--lr", default=LR)
    parser.add_argument("--optim", default=OPTIM)
    args = parser.parse_args()
    
    NUM_SEEDS = int(args.n_seed)
    ACTF_FN = args.actf_fn
    LR = float(args.lr)
    OPTIM = args.optim

    loss_matrix = []
    reward_matrix = []

    for SEED in range(1, NUM_SEEDS+1):

        env = gym.make(ENV_ID)

        torch.manual_seed(SEED)
        np.random.seed(SEED)
        random.seed(SEED)
        env.action_space.seed(SEED)
        env.seed(SEED)

        current_model = DQN(env.observation_space.shape[0], env.action_space.n, ACTF_FN).to(device)
        target_model  = DQN(env.observation_space.shape[0], env.action_space.n, ACTF_FN).to(device)
            
        optimizer = None
        if OPTIM == 'Adam':
            optimizer = optim.Adam(current_model.parameters(), lr=LR)
        elif OPTIM == 'SGD':
            optimizer = optim.SGD(current_model.parameters(), lr=LR)
        else:
            raise ValueError("Bad argument for optim")

        replay_buffer = ReplayBuffer(5000)

        update_target(current_model, target_model)

        losses = []
        all_rewards = []
        episode_reward = 0
        episode = 1
        frame_idx = 0

        state = env.reset()
        while episode <= NUM_EPISODES:
            frame_idx += 1
            epsilon = EPSILON_FINAL + (EPSILON_START - EPSILON_FINAL) * np.exp(-1. * frame_idx / EPSILON_DECAY)
            action = current_model.act(state, epsilon)
            
            next_state, reward, done, _ = env.step(action)
            replay_buffer.push(state, action, reward, next_state, done)
            
            state = next_state
            episode_reward += reward
            
            if done:
                state = env.reset()
                all_rewards.append(episode_reward)
                episode_reward = 0
                episode += 1
                
            if len(replay_buffer) > BATCH_SIZE:
                loss = compute_td_loss(BATCH_SIZE)
                losses.append(loss.item())
                
            if frame_idx % 100 == 0:
                update_target(current_model, target_model)

        print("Seed {} complete".format(SEED))
        loss_matrix.append(losses)
        reward_matrix.append(all_rewards)

    #plot(loss_matrix, reward_matrix, error_bars=False)
    plot2(reward_matrix)
    save_data(reward_matrix)