import math, random

import gym
import numpy as np

import sys
import torch
import torch.nn as nn
import torch.optim as optim
import torch.autograd as autograd 
import torch.nn.functional as F

import sys
sys.path.append("..")

import general_energy_nets.predictive_coding as pc

from IPython.display import clear_output
import matplotlib.pyplot as plt

from collections import deque

import argparse

assert torch.cuda.is_available()
device = torch.device("cuda" if torch.cuda.is_available() else "cpu" )

ENV_ID = "CartPole-v1"
NUM_SEEDS = 10

EPSILON_START = 1.0
EPSILON_FINAL = 0.01
EPSILON_DECAY = 500

NUM_EPISODES = 100
BATCH_SIZE = 64
GAMMA = 0.98

BIG_T = 64
X_LR = 0.1
ACTF_FN = 'ReLU'
UPDATE_P = 'last'
P_OPTIM = 'Adam'
CLAMP = False
USE_PC_PLOTTER = False

class ReplayBuffer(object):
    def __init__(self, capacity):
        self.buffer = deque(maxlen=capacity)
    
    def push(self, state, action, reward, next_state, done):
        state      = np.expand_dims(state, 0)
        next_state = np.expand_dims(next_state, 0)
            
        self.buffer.append((state, action, reward, next_state, done))
    
    def sample(self, batch_size):
        state, action, reward, next_state, done = zip(*random.sample(self.buffer, batch_size))
        return np.concatenate(state), action, reward, np.concatenate(next_state), done
    
    def __len__(self):
        return len(self.buffer)

class PCDQN(nn.Module):
    def __init__(self, num_inputs, num_actions, actf_fn):
        super(PCDQN, self).__init__()
        
        #TODO allow for other activation functions to be specified

        self.lin1 = nn.Linear(env.observation_space.shape[0], 128)
        self.lin2 = nn.Linear(128, 128)
        self.lin3 = nn.Linear(128, env.action_space.n)
        self.pc1 = pc.PCLayer()
        self.pc2 = pc.PCLayer()
           
    def pc_layers(self):
        return [self.pc1, self.pc2]
    
    def forward(self, x):
        x = self.lin1(x)
        x = self.pc1(x)
        x = F.relu(x)
        x = self.lin2(x)
        x = self.pc2(x)
        x = F.relu(x)
        return self.lin3(x)
    
    @torch.no_grad()
    def act(self, state, epsilon):
        self.train(False)
        if random.random() > epsilon:
            state   = torch.FloatTensor(state).unsqueeze(0).to(device)
            q_value = self.forward(state)
            action  = q_value.max(1)[1].data[0].item()
        else:
            action = random.randrange(env.action_space.n)
        return action

def loss_fn(output, _expected_q_value, _action):
    q_value = output.gather(1, _action.unsqueeze(1)).squeeze(1) 
    return (q_value - _expected_q_value).pow(2).sum() * 0.5

def clamp_loss_fn(output, _targets):
    return (output - _targets).pow(2).sum() * 0.5

def update_target(current_model, target_model):
    parameters = current_model.state_dict()
    try:
        del parameters['pc1._x']
        del parameters['pc2._x']
    except:
        pass
    target_model.load_state_dict(parameters)

def get_energy_at_layers(t):
    with torch.no_grad():
        energy_at_layers = [torch.clone(pc_layer.energy()).detach().cpu().sum().item() for pc_layer in current_model.pc_layers()]
        pc_plotter.append_energy_at_layers(energy_at_layers)

def compute_td_loss(batch_size):
    state, action, reward, next_state, done = replay_buffer.sample(batch_size)

    state      = torch.FloatTensor(np.float32(state)).to(device)
    next_state = torch.FloatTensor(np.float32(next_state)).to(device)
    action     = torch.LongTensor(action).to(device)
    reward     = torch.FloatTensor(reward).to(device)
    done       = torch.BoolTensor(done).to(device)

    #q_values      = current_model(state)
    #q_value       = q_values.gather(1, action.unsqueeze(1)).squeeze(1) 
    
    current_model.eval()
    
    with torch.no_grad():
        next_q_values = current_model(next_state)
        next_q_state_values = target_model(next_state) 
        next_q_value = next_q_state_values.gather(1, torch.max(next_q_values, 1)[1].unsqueeze(1)).squeeze(1)
        next_q_value[done] = 0.0
        next_q_value = next_q_value.detach()

    expected_q_value = reward + GAMMA * next_q_value
    
    current_model.train()
    results = pc_trainer.train_on_batch(
        inputs=state,
        loss_fn=loss_fn,
        loss_fn_kwargs={
            '_expected_q_value': expected_q_value,
            '_action': action
        },
        is_log_progress=False,
        is_return_results_every_t=USE_PC_PLOTTER,
        callback_after_t=get_energy_at_layers if USE_PC_PLOTTER else None,
        is_checking_after_callback_after_t=False,
    )
    
    return results

def compute_td_loss_with_clamp(batch_size):
    state, action, reward, next_state, done = replay_buffer.sample(batch_size)

    state      = torch.FloatTensor(np.float32(state)).to(device)
    next_state = torch.FloatTensor(np.float32(next_state)).to(device)
    action     = torch.LongTensor(action).to(device)
    reward     = torch.FloatTensor(reward).to(device)
    done       = torch.BoolTensor(done).to(device)

    #q_values      = current_model(state)
    #q_value       = q_values.gather(1, action.unsqueeze(1)).squeeze(1) 
    
    current_model.eval()
    
    with torch.no_grad():
        next_q_values = current_model(next_state)
        next_q_state_values = target_model(next_state) 
        next_q_value = next_q_state_values.gather(1, torch.max(next_q_values, 1)[1].unsqueeze(1)).squeeze(1)
        next_q_value[done] = 0.0
        next_q_value = next_q_value.detach()
        
        current_q_values = current_model(state)
        current_q_values = current_q_values.detach()
        
        batch_index = torch.arange(current_q_values.shape[0]).to(device)
        
    
    expected_q_value = reward + GAMMA * next_q_value
    current_q_values[batch_index, action] = expected_q_value
    
    current_model.train()
    results = pc_trainer.train_on_batch(
        inputs=state,
        loss_fn=clamp_loss_fn,
        loss_fn_kwargs={
            '_targets': current_q_values
        },
        is_log_progress=False,
        is_return_results_every_t=USE_PC_PLOTTER,
        callback_after_t=get_energy_at_layers if USE_PC_PLOTTER else None,
        is_checking_after_callback_after_t=False,
    )
    
    return results

class PC_Plotter:
    
    def __init__(self, plot_infer_every=1000):
        
        self.losses_before_infer = []
        self.losses_after_infer = []
        self.energy_after_infer = []
        
        self.energy_at_layers = []
        
        self.plot_infer_every = plot_infer_every
        self._seed = 1
    
    def set_seed(self, seed):
        self._seed = seed

    def append_results(self, results, idx):
        
        self.losses_before_infer.append(results['loss'][0])
        self.losses_after_infer.append(results['loss'][-1])
        
        self.energy_after_infer.append(self.energy_at_layers[-1].copy())
        
        if idx == (BATCH_SIZE+1) or idx % self.plot_infer_every == 0:
            self.plot_infer(results['loss'], self.energy_at_layers, idx)
            
        self.energy_at_layers = []
            
    def append_energy_at_layers(self, energy_at_layers):
        self.energy_at_layers.append(energy_at_layers.copy())
        
    def plot_infer(self, loss, energy_at_layers, idx):
        energy_at_layers = np.array(energy_at_layers, dtype=np.float32)
        plt.style.use('ggplot')
        plt.figure(figsize=(20,15))
        
        #plot loss during inference
        plt.subplot(231)
        plt.title('Loss during inference')
        plt.plot(loss, linewidth=1, color='red', alpha=0.8)
        
        #plot total energy during inference
        plt.subplot(232)
        plt.title('Total enegry during inference')
        total_energy = energy_at_layers.sum(axis=1)
        plt.plot(total_energy, linewidth=1, color='green', alpha=0.8)
        
        #plot energy at layer one during inference
        plt.subplot(233)
        plt.title('Energy at pc1 during inference')
        plt.plot(energy_at_layers[:, 0], linewidth=1, color='orange', alpha=0.8)
        
        #plot enegry at layer two during inference
        plt.subplot(234)
        plt.title('Energy at pc2 during inference')
        plt.plot(energy_at_layers[:, 1], linewidth=1, color='blue', alpha=0.8)
        
        #plot total during inference
        plt.subplot(235)
        plt.title('Total loss and energy during inference')
        total = loss + total_energy
        plt.plot(total, linewidth=1, color='purple', alpha=0.8)
        
        #plot all of them on one axis
        plt.subplot(236)
        plt.plot(loss, linewidth=1, color='red', alpha=0.8, label="loss")
        plt.plot(total_energy, linewidth=1, color='green', alpha=0.8, label="total energy")
        plt.plot(energy_at_layers[:, 0], linewidth=1, color='orange', alpha=0.8, label="energy at pc1")
        plt.plot(energy_at_layers[:, 1], linewidth=1, color='orange', alpha=0.8, label="energy at pc2")
        plt.plot(total, linewidth=1, color='purple', alpha=0.8, label="total")
        plt.legend()
        plt.savefig('./pc_plotter/seed{}/inference_plot_frame_idx={}.png'.format(self._seed, idx))

        plt.close('all')
        
    def plot_final(self):
        energy_after_infer = np.array(self.energy_after_infer, dtype=np.float32) 
        
        #plot loss before inference
        plt.style.use('ggplot')
        plt.figure(figsize=(10,5))
        plt.title('Loss before inference')
        plt.plot(self.losses_before_infer, linewidth=2, color='blueviolet', alpha=0.8)
        plt.savefig('./pc_plotter/seed{}/loss_before_inference.png'.format(self._seed))
        
        #plot loss after inference
        plt.style.use('ggplot')
        plt.figure(figsize=(10,5))
        plt.title('Loss after inference')
        plt.plot(self.losses_after_infer, linewidth=2, color='firebrick', alpha=0.8)
        plt.savefig('./pc_plotter/seed{}/loss_after_inference.png'.format(self._seed))
        
        #plot total energy after inference
        plt.style.use('ggplot')
        plt.figure(figsize=(10,5))
        plt.title('Total energy after inference')
        total_energy_after_inf = energy_after_infer.sum(axis=1)
        plt.plot(total_energy_after_inf, linewidth=2, color='forestgreen', alpha=0.8)
        plt.savefig('./pc_plotter/seed{}/total_energy_after_inference.png'.format(self._seed))
        
        #plot energy at layer one after inference
        plt.style.use('ggplot')
        plt.figure(figsize=(10,5))
        plt.title('Energy at pc1 after inference')
        plt.plot(energy_after_infer[:, 0], linewidth=2, color='gold', alpha=0.8)
        plt.savefig('./pc_plotter/seed{}/energy_pc1_after_inference.png'.format(self._seed))
        
        #plot energy at layer two after inference
        plt.style.use('ggplot')
        plt.figure(figsize=(10,5))
        plt.title('Energy at pc2 after inference')
        plt.plot(energy_after_infer[:, 1], linewidth=2, color='lawngreen', alpha=0.8)
        plt.savefig('./pc_plotter/seed{}/energy_pc2_after_inference.png'.format(self._seed))
        
        #plot total loss + energy after inference
        plt.style.use('ggplot')
        plt.figure(figsize=(10,5))
        plt.title('Total loss and energy after inference')
        total = self.losses_after_infer + total_energy_after_inf
        plt.plot(total, linewidth=2, color='violet', alpha=0.8)
        plt.savefig('./pc_plotter/seed{}/total_loss_and_energy_after_inference.png'.format(self._seed))

        plt.close('all')
        
def plot(loss_matrix, reward_matrix, error_bars=False):
    plt.style.use('ggplot')
    loss_matrix = np.array(loss_matrix)
    min_episodes = min([len(rewards) for rewards in reward_matrix])
    reward_matrix = np.array([rewards[:min_episodes] for rewards in reward_matrix])

    plt.figure(figsize=(20,5))
    plt.subplot(131)
    plt.title('rewards')

    reward_matrix_mean = np.mean(reward_matrix, axis=0)
    reward_matrix_percentiles = np.percentile(reward_matrix, [5, 25, 50, 75, 95], axis=0)

    plt.plot(reward_matrix_mean, linewidth=2, color='green', alpha=0.8)
    if error_bars:
        plt.fill_between([x for x in range(reward_matrix.shape[1])], reward_matrix_percentiles[1, :], reward_matrix_percentiles[3, :], color='#888888', alpha=0.4)
        plt.fill_between([x for x in range(reward_matrix.shape[1])], reward_matrix_percentiles[0, :], reward_matrix_percentiles[4, :], color='#888888', alpha=0.2)
    else:
        for i in range(reward_matrix.shape[0]):
            plt.plot(reward_matrix[i, :], color='gray', linestyle='dashed', alpha=0.5)
    
    plt.subplot(132)
    plt.title('loss')

    loss_matrix_mean = np.mean(loss_matrix, axis=0)
    loss_matrix_percentiles = np.percentile(loss_matrix, [5, 25, 50, 75, 95], axis=0)

    plt.plot(loss_matrix_mean, linewidth=2, alpha=0.8)
    plt.fill_between([x for x in range(loss_matrix.shape[1])], loss_matrix_percentiles[1, :], loss_matrix_percentiles[3, :], color='#888888', alpha=0.4)
    plt.fill_between([x for x in range(loss_matrix.shape[1])], loss_matrix_percentiles[0, :], loss_matrix_percentiles[4, :], color='#888888', alpha=0.2)

    plt.savefig('./imgs/dqn_pc_seeds={}_T={}_x_lr={:.2f}_actf_fn={}_p_optim={}_update_p={}_clamp={}.png'.format(NUM_SEEDS, str(BIG_T), X_LR, ACTF_FN, P_OPTIM, UPDATE_P, CLAMP))
    plt.close('all')


def plot2(reward_matrix):
    plt.style.use('ggplot')
    max_average_reward = np.max(np.mean(reward_matrix, axis=0))
    plt.figure(figsize=(10, 5))
    plt.title('CartPole-v1 : Vanilla DQN with PC-C (Max Avg Reward = {:.2f})'.format(max_average_reward))
    reward_matrix_mean = np.mean(reward_matrix, axis=0)
    plt.plot(reward_matrix_mean, linewidth=2, color='firebrick', alpha=0.8)

    upper_confidence_interval = []
    lower_confidence_interval = []

    for episode in reward_matrix.transpose():
        bootstrapped_means = []
        for B in range(2000):
            bootstrapped_sample = np.random.choice(episode, len(episode))
            bootstrapped_means.append(np.mean(bootstrapped_sample))
        bootstrapped_percentiles = np.percentile(np.array(bootstrapped_means), [2.5, 97.5])
        upper_confidence_interval.append(bootstrapped_percentiles[1])
        lower_confidence_interval.append(bootstrapped_percentiles[0])

    plt.ylim([0, 500])
    plt.xlabel('episodes')
    plt.ylabel('reward')
    plt.fill_between([x for x in range(reward_matrix.shape[1])], upper_confidence_interval, lower_confidence_interval, color='firebrick', alpha=0.4)
    plt.savefig('./imgs/dqn_pc_rewards_seeds={}_T={}_x_lr={:.2f}_actf_fn={}_p_optim={}_update_p={}_clamp={}.png'.format(NUM_SEEDS, str(BIG_T), X_LR, ACTF_FN, P_OPTIM, UPDATE_P, CLAMP))
    plt.close('all')


def save_data(reward_matrix):
    max_average_reward = np.max(np.mean(reward_matrix, axis=0))
    with open('./data/dqn_pc_seeds={}_T={}_x_lr={:.2f}_actf_fn={}_p_optim={}_update_p={}_clamp={}.txt'.format(NUM_SEEDS, str(BIG_T), X_LR, ACTF_FN, P_OPTIM, UPDATE_P, CLAMP), 'w') as outfile:
        outfile.write('Max Average Reward : {:.2f}'.format(max_average_reward))
    print('T : {}, x_lr : {:.2f}, actf_fn : {}, p_optim : {}, update_p : {}, clamp : {}'.format(str(BIG_T), X_LR, ACTF_FN, P_OPTIM, UPDATE_P, CLAMP))
    print('Experiment done : Max Average Reward : {:.2f}'.format(max_average_reward))

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--n-seed", default=NUM_SEEDS)
    parser.add_argument("--T", default=BIG_T)
    parser.add_argument("--x-lr", default=X_LR)
    parser.add_argument("--actf-fn", default=ACTF_FN)
    parser.add_argument("--p-optim", default=P_OPTIM)
    parser.add_argument("--update-p", default=UPDATE_P)
    parser.add_argument("--clamp", default=CLAMP)
    parser.add_argument("--pc-plotter", default=USE_PC_PLOTTER)
    args = parser.parse_args()
    
    NUM_SEEDS = int(args.n_seed)
    BIG_T = int(args.T)
    X_LR = float(args.x_lr)
    ACTF_FN = args.actf_fn
    P_OPTIM = args.p_optim
    UPDATE_P = args.update_p
    CLAMP = bool(int(args.clamp))
    USE_PC_PLOTTER = bool(int(args.pc_plotter))
    
    pc_plotter = PC_Plotter() if USE_PC_PLOTTER else None

    loss_matrix = []
    reward_matrix = []

    for SEED in range(1, NUM_SEEDS+1):

        env = gym.make(ENV_ID)

        torch.manual_seed(SEED)
        np.random.seed(SEED)
        random.seed(SEED)
        env.action_space.seed(SEED)
        env.seed(SEED)

        current_model = PCDQN(env.observation_space.shape[0], env.action_space.n, ACTF_FN).to(device)
        target_model  = PCDQN(env.observation_space.shape[0], env.action_space.n, ACTF_FN).to(device)
        target_model.eval()

        replay_buffer = ReplayBuffer(5000)

        p_optim = optim.Adam

        if P_OPTIM == "Adam":
            p_optim = optim.Adam
        elif P_OPTIM == "SGD":
            p_optim = optim.SGD
        else:
            raise ValueError("Bad argument for p-optim")

        pc_trainer = pc.PCTrainer(
            current_model,
            T=BIG_T,
            update_x_at='all',
            optimizer_x_fn=optim.SGD,
            optimizer_x_kwargs={"lr": X_LR},
            x_lr_discount=0.5,
            update_p_at=UPDATE_P,
            optimizer_p_fn=p_optim,
            optimizer_p_kwargs={"lr": 0.001},
            plot_progress_at=[],
        )

        update_target(current_model, target_model)

        if USE_PC_PLOTTER:
            pc_plotter.set_seed(SEED)

        losses = []
        all_rewards = []
        episode_reward = 0
        episode = 1
        frame_idx = 0

        state = env.reset()
        while episode <= NUM_EPISODES:
            frame_idx += 1
            epsilon = EPSILON_FINAL + (EPSILON_START - EPSILON_FINAL) * np.exp(-1. * frame_idx / EPSILON_DECAY)
            action = current_model.act(state, epsilon)
            
            next_state, reward, done, _ = env.step(action)
            replay_buffer.push(state, action, reward, next_state, done)
            
            state = next_state
            episode_reward += reward
            
            if done:
                state = env.reset()
                all_rewards.append(episode_reward)
                episode_reward = 0
                episode += 1
                
            if len(replay_buffer) > BATCH_SIZE:
                if CLAMP:
                    results = compute_td_loss_with_clamp(BATCH_SIZE)
                else:
                    results = compute_td_loss(BATCH_SIZE)
                losses.append(results['loss'][-1])

                if USE_PC_PLOTTER:
                    pc_plotter.append_results(results, frame_idx)
                
            if frame_idx % 100 == 0:
                update_target(current_model, target_model)

        if USE_PC_PLOTTER:
            pc_plotter.plot_final()
        
        print("Seed {} complete".format(SEED))
        loss_matrix.append(losses)
        reward_matrix.append(all_rewards)

    #plot(loss_matrix, reward_matrix, error_bars=False)
    plot2(reward_matrix)
    save_data(reward_matrix)
    

    
    
    